# Now this file becomes Lattice1.0.mpl

Lattice:=module()
option `Copyright 2016 H.-Ulrich (Uli) Wienands. All Rights Reserved.`,package;
description "First-order optics, tracking and export to lattice programs.";
export J,Drift,Bend,Quad,QuadOld,Sextupole,Solenoid,RfCavity,GKick,Wire,Foil;
export DefineElement,DefineLine,DefineMachine,DefineBeam,ExpandLine,LumpLine;
export `.`,`^`,Subs,SRotate;
export TwissTran,EtaTran,HxTran,Tunes,Track;
export cosmux,cosmuy,sinmux,sinmuy,nux,nuy;
export betax,alphax,gammax,betay,alphay,gammay,twissx,twissy,twiss;
export etax,etapx,etay,etapy,dispersion,Hx,Hy,H;
export Mad8Form,BeamPlot,LatticePlot;
export I1,I2,I3,I4x,I5x;
export ST:=Vector(1,0,datatype=float[8]);
export Edefault:=0;
export dET:=Vector(1,0,datatype=float[8]);
export ElnamT:=Array(datatype=anything); # Can't be string! datatype=string fails for more than a few elements
export betaxT:=Vector(1,0,datatype=float[8]);
export alphaxT:=Vector(1,0,datatype=float[8]);
export etaxT:=Vector(1,0,datatype=float[8]);
export etapxT:=Vector(1,0,datatype=float[8]);
export betayT:=Vector(1,0,datatype=float[8]);
export alphayT:=Vector(1,0,datatype=float[8]);
export etayT:=Vector(1,0,datatype=float[8]);
export etapyT:=Vector(1,0,datatype=float[8]);

local ModuleLoad,ModuleUnLoad;

# Some auxiliary constructs; local to the Lattice module

local Tquad:=(k) -> Matrix([[1,0,0,0,0,0],[-k,1,0,0,0,0],[0,0,1,0,0,0],[0,0,k,1,0,0],[0,0,0,0,1,0],[0,0,0,0,0,1]]);

# The tables for the output (...Form-) commands:
local DriftTable:=table();
local QuadTable:=table();
local BendTable:=table();
local LineTable:=table();
local Driftidx:=0,Quadidx:=0,Bendidx:=0,Lineidx:=0,recursions:=0;
local Ea,sAct,index;
local OutTables,lattab,expand_line,trackElement;
local TwissT:=Vector(6,0);
local EtaT:=Vector(4,0);
local ParticleTable:=table();
local Elems:=Vector(1,0,datatype=anything); # for reasons I don't get this can't be an array.
local BeamParticles:=0; # no. of particles of last DefineBeam;
local BeamEnergy; # set by last DefineBeam;
local BeamGammar; # set by last DefineBeam;
local BeamBetar; # set by last DefineBeam;
local BeamParticleRestMass; # set by last DefineBeam;
local typesettingOld;

    (* Functions describing the integrals of curly Hx and of etax through elements with bending radius rho.
   These are considered correct as of 11-Mar-2015
 *)
local HIntf:=(tw,disp,rho,Theta) ->
    rho*(2*disp[2]-2*cos(Theta)*disp[2]+Theta*disp[2]^2-(1/2)*sin(Theta)*cos(Theta)+(1/2)*Theta)*tw[1]+((2*sin(Theta)*disp[2]-cos(Theta)^2-2*Theta*disp[2]+2*cos(Theta)-1)*rho^2+(2*Theta*disp[1]*disp[2]-2*cos(Theta)*disp[1]+2*disp[1])*rho)*tw[2]+((((1/2)*sin(Theta)*cos(Theta)-2*sin(Theta)+(3/2)*Theta)*rho^3+(2*sin(Theta)*disp[1]-2*Theta*disp[1])*rho^2+Theta*disp[1]^2*rho)*tw[2]^2+((1/2)*sin(Theta)*cos(Theta)-2*sin(Theta)+(3/2)*Theta)*rho^3+(2*sin(Theta)*disp[1]-2*Theta*disp[1])*rho^2+Theta*disp[1]^2*rho)/tw[1]:

local dispIntf:=(disp,rho,Theta) ->
    rho*(disp[2]*rho+sin(Theta)*disp[1]-cos(Theta)*rho*disp[2]+Theta*rho-sin(Theta)*rho):

# The multiple-scattering formula used in Foil and Collimator. In MeV, m and rad
local MSTh0:=(X0,x,p) -> 13.6/p*sqrt(x/X0)*(1+0.038*ln(x/X0));

    unprotect(Spin);
    unprotect('Elems');
    unprotect('Edefault');

    unprotect('betaxT');
    unprotect('alphaxT');
    unprotect('etaxT');
    unprotect('etapxT');
    unprotect('betayT');
    unprotect('alphayT');
    unprotect('etayT');
    unprotect('etapyT');

# Our prototypes: (debugged & working)

#------------------ Symplectic Matrix J --------------------------
    J:=<<0,-1,0,0,0,0>|<1,0,0,0,0,0>|<0,0,0,-1,0,0>|<0,0,1,0,0,0>|<0,0,0,0,0,-1>|<0,0,0,0,1,0>>;

#######################################################################
#--------- Element Definitions --------------------------------------
#######################################################################
#--------- Drift --------------------------------------
    Drift:=proc(len,Er:=0)
    description "Drift(len,Eref) defined a drift element. Eref is optional.";
    local Rmat:=Matrix(6,6,\
                       {(1,2)=len,(3,4)=len,(1,1)=1,(2,2)=1,(3,3)=1,(4,4)=1,(5,5)=1,(6,6)=1},\
                       fill=0);
    local trackFunction:= (particleVector) -> LinearAlgebra:-MatrixVectorMultiply(Rmat,particleVector);
        return Record('l'=len,'kind'=cat('Drift'),'dE'=0,'Eref'=Er,\
                      'R'=Rmat,'TF'=trackFunction);
    end proc; # Drift

#--------- Quad --------------------------------------

    Quad:=proc(len,kl,n:=0,Er:=0)
    description "Quad(len,kl,nslices,Eref) defines a quadrupole. nslices=1 for thin quad, 0 or infinity for thick quad. nslices is optional, Eref is optional.";
    local Rmat, trackFunction,nsl;

        if ((n=0 or n=infinity) and len<>0) then # thick quad

            if (kl<>0) then
                Rmat:=Matrix(6,6,\
                             {(1, 1) = cos(len^(1/2)*kl^(1/2)),\
                              (1, 2) = sin(len^(1/2)*kl^(1/2))*len^(1/2)/kl^(1/2),\
                              (2, 1) = -kl^(1/2)*sin(len^(1/2)*kl^(1/2))/len^(1/2),\
                              (2, 2) = cos(len^(1/2)*kl^(1/2)),\
                              (3, 3) = cosh((len*kl)^(1/2)),\
                              (3, 4) = sinh((len*kl)^(1/2))/kl^(1/2)*len^(1/2),\
                              (4, 3) = kl^(1/2)*sinh(len^(1/2)*kl^(1/2))/len^(1/2),\
                              (4, 4) = cosh(len^(1/2)*kl^(1/2)),\
                              (5,5)=1,(6,6)=1},fill=0)
            else
                Rmat:=Matrix(6,6,\
                             {(1, 1) = 1,\
                              (1, 2) = len,\
                              (2, 1) = 0,\
                              (2, 2) = 1,\
                              (3, 3) = 1,\
                              (3, 4) = len,\
                              (4, 3) = 0,\
                              (4, 4) = 1,\
                              (5,5)=1,(6,6)=1},fill=0)
            end if;

            trackFunction:= (particleVector) -> LinearAlgebra:-MatrixVectorMultiply(Rmat,particleVector);
            return Record('l'=len,'k1l'=kl,'ns'=n,'Eref'=Er,'dE'=0,'kind'=cat('Quad'),\
                          'R'=Rmat,'TF'=trackFunction);
        else # sliced quad
            nsl:=max(n,1); # can't have 0 slices
            Rmat:=LinearAlgebra:-MatrixPower(LinearAlgebra:-Multiply(Drift(len/(2*nsl)):-R,\
                                                                     LinearAlgebra:-Multiply(Tquad(kl/nsl),\
                                                                                             Drift(len/(2*nsl)):-R)),\
                                             nsl);
            trackFunction:= (particleVector) -> LinearAlgebra:-MatrixVectorMultiply(Rmat,particleVector);
            return Record('l'=len,'k1l'=kl,'ns'=nsl,'Eref'=Er,'dE'=0,'kind'=cat('Quad'),\
                          'R'=Rmat,'TF'=trackFunction);
        end if; # thin or thick quad

    end proc; # Quad

#--------- Solenoid --------------------------------------
    Solenoid:=proc(len,k)
    local ks;
    local Rmat,trackFunction;

        if (type(k,extended_numeric)) then
            Rmat:=limit~(Matrix(6,6,\
                                {(1,1)=cos(ks*len)^2,(1,2)=sin(ks*len)*cos(ks*len)/ks,\
                                 (1,3)=sin(ks*len)*cos(ks*len),\
                                 (1,4)=sin(ks*len)^2/ks,\
                                 (2,1)=-ks*sin(ks*len)*cos(ks*len),(2,2)=cos(ks*len)^2,
                                 (2,3)=-ks*sin(ks*len)^2,(2,4)=sin(ks*len)*cos(ks*len),\
                                 (3,1)=-sin(ks*len)*cos(ks*len),(3,2)=-sin(ks*len)^2/ks,\
                                 (3,3)=cos(ks*len)^2,(3,4)=sin(ks*len)*cos(ks*len)/ks,\
                                 (4,1)=ks*sin(ks*len)^2,(4,2)=-sin(ks*len)*cos(ks*len),\
                                 (4,3)=-ks*sin(ks*len)*cos(ks*len),(4,4)=cos(ks*len)^2,\
                                 (5,5)=1,(6,6)=1},fill=0),ks=k);

        else
            ks:=k;
            Rmat:=Matrix(6,6,\
                         {(1,1)=cos(ks*len)^2,(1,2)=sin(ks*len)*cos(ks*len)/ks,\
                          (1,3)=sin(ks*len)*cos(ks*len),\
                          (1,4)=sin(ks*len)^2/ks,\
                          (2,1)=-ks*sin(ks*len)*cos(ks*len),(2,2)=cos(ks*len)^2,
                          (2,3)=-ks*sin(ks*len)^2,(2,4)=sin(ks*len)*cos(ks*len),\
                          (3,1)=-sin(ks*len)*cos(ks*len),(3,2)=-sin(ks*len)^2/ks,\
                          (3,3)=cos(ks*len)^2,(3,4)=sin(ks*len)*cos(ks*len)/ks,\
                          (4,1)=ks*sin(ks*len)^2,(4,2)=-sin(ks*len)*cos(ks*len),\
                          (4,3)=-ks*sin(ks*len)*cos(ks*len),(4,4)=cos(ks*len)^2,\
                          (5,5)=1,(6,6)=1},fill=0);
        end if;

        trackFunction:= (particleVector) -> LinearAlgebra:-MatrixVectorMultiply(Rmat,particleVector);
        return Record('l'=len,'a'=a1,'Eref'=Er,'kind'=cat('Solenoid'),'dE'=((E) -> 0),\
                      'R'=Rmat,'TF'=trackFunction);

    end proc; # solenoid

#--------- Bend --------------------------------------
    Bend:=proc(len,a1,n1,Er:=0)
    description "Bend(len,angle,n,Eref) defines an Sbend. len is the path length. n is the field index. Eref is optional.";
# Checked against DIMAD SBEND and agrees to machine precision. Be careful with large n, though, which may be numerically unstable
# NOTE: Changed sign or R[5,1] and R[5,2] as this thing was not symplectic!
# This is OPPOSITE to DIMAD! It does agree with Wolski's definition as used in his lecture series.
# At the root here is the use of non-canonical variables.
# (Det(R) ≠ 1).
    local h:=a1/len;
# local rho;
    local kx:=sqrt(1-Bend_n)*h;
    local ky:=sqrt(Bend_n)*h;
    local V_0:=-8.84627577501730E-14*Er^4*h * a1/(2*Pi);
    local Rmat:=limit~(Matrix(6,6,\
                              {(1,1)=cos(kx*len),(1,2)=sin(kx*len)/kx,(1,6)=h/(kx*kx)*(1-cos(kx*len)),\
                               (2,1)=-kx*sin(kx*len),(2,2)=cos(kx*len),(2,6)=h/kx*sin(kx*len),\
                               (3,3)=cos(ky*len),(3,4)=sin(ky*len)/ky,\
                               (4,3)=-ky*sin(ky*len),(4,4)=cos(ky*len),\
                               (5,1)=-h/kx*sin(kx*len),(5,2)=-h/(kx*kx)*(1-cos(kx*len)),(5,5)=1,\
                               (5,6)=h*h/(kx*kx*kx)*(kx*len-sin(kx*len)),\
                               (6,6)=1},fill=0),Bend_n=n1);
    local trackFunction:= (particleVector) -> LinearAlgebra:-MatrixVectorMultiply(Rmat,particleVector);
        return Record('l'=len,'a'=a1,'E1'=0,'E2'=0,'n'=n1,'Eref'=Er,'kind'=cat('Bend'),'rho'=1/h,\
                      'dE'=((E) -> -8.84627577501730E-14*E^4*(self:-a/self:-l) * self:-a/(2*Pi)),\
                      'R'=Rmat,'TF'=trackFunction);
    end proc; # Bend

#--------- Sextupole --------------------------------------
    Sextupole:=proc(len,k2L)
    description "Implements a thin sextupole.";
    local k2len:=k2L;
    local Trf:=Drift(len/2):-TF; # We track like a drift of length len/2.
    local Rmat:=Drift(len):-R; # to 1st order, it is a drift
    local trackFunction;

        trackFunction:=proc(pV)
        local rV:=Trf(pV); # track in by 1/2 length
            rV:=rV+~<0,k2len*(rV[3]^2-rV[1]^2),0,k2len*2*rV[1]*rV[3],0,0>;
            rV:=Trf(rV);
            return rV;
        end proc;

        return Record('l'=len,'k2l'=k2len,'Eref'=0,'kind'=cat('Sextupole'),'dE'=0,'R'=Rmat,\
                      'TF'=trackFunction);
    end proc; # Sextupole

#--------- RfCavity --------------------------------------
    RfCavity:=proc(L,freq,V0,phi)
# Correct longitudinal phase space for phi=0 demonstrated Sept. 10 using USPAS toy ring.
    description "Implements an rf cavity. Units are m, Hz, MV, and rad";
    local clight:=299792458;
    local len:=L; # [m]
    local phi_rf:=phi;
    local sphi:=sin(phi);
    local cphi:=cos(phi);
    local Vcav:=V0;
    local rfPeriod:=1/freq; # [1/sec]
    local BeamVelocity:=BeamBetar*clight; # [m/s]
    local time_delta:=len/(2*BeamVelocity);
    local Rmat:=Drift(len):-R; # cavity is just a drift to 1st order.
    local Trf:=Drift(len/2):-TF; # for track-in and track-out.

    local trackFunction:=proc(pV)
          local time_delta:=pV[5]/(2*BeamBetar*clight); # fraction of a period
          local phase_delta:=time_delta/rfPeriod*2*Pi; # in [2Pi]
          local dE:=Vcav*sin(phi_rf+phase_delta);
          local rV:=Trf(pV); # track in by 1/2 length
              rV[6]:=rV[6]+dE/BeamEnergy; # apply longitudinal kick
              return Trf(rV); # Track out 1/2 length and return
          end proc; # trackFunction

        return Record('l'=L,'kind'=cat('RfCavity'),'R'=Rmat,\
                      'TF'=trackFunction,'Eref', dE=((E) -> Vcav*sphi));
    end proc; # RfCavity

#--------- GKick --------------------------------------
    GKick:=proc(dxk:=0,dxpk:=0,dyk:=0,dypk:=0,dlk:=0,ddpk:=0)
    description "Implements a general kick with 0 length. Units are m, radians and frac. momentum";

    local trackFunction:= (particleVector) -> particleVector+~Vector(6,\
                                                                     [dxk,dxpk,dyk,dypk,dlk,ddpk]);

        return Record('l'=0,'kind'=cat('GKick'),'R'=LinearAlgebra:-IdentityMatrix(6),\
                      'dx'=dxk,'dxp'=dxpk,'dy'=dyk,'dyp'=dypk,'dl'=dlk,'ddp'=ddpk,\
                      'dE'=0,'Eref'=0,'TF'=trackFunction);

    end proc; # Gkick

#--------- Wire --------------------------------------
    (* Wire(len,current,xsep,ysep,Eref:=0)
   Implements a wire parallel to the beam; e.g. as beam-beam corrector
   Needs the beam energy to work
 *)
    Wire:=proc(len,current,xsep,ysep,Ener:=0)
    local Brho:=`if`(Ener<>0,3.33564E-3*Ener,'Brho'); # Note slight approximation
    local Bb:=Vector(2, {(1) = y*mu_0*current/(2*Pi*(x^2+y^2)),\
                         (2) = x*mu_0*current/(2*Pi*(x^2+y^2))}); # [1] is Bx; [2] is By
    local kick:=[seq(mtaylor(Bb[ii],[x=-xsep,y=-ysep],4),ii=1..2)]\
        *~len/~Brho; # [1] is ky; [2] is kx
    local kickf:=unapply(kick,[x,y]);
    local k:=coeff(simplify(mtaylor(Bb[2],[x=-xsep,y=-ysep],2)),x)/Brho; # ∂By/∂x / Brho
    local trackFunction:=(pV) -> <pV[1],pV[2]+kickf(pV[1]-xsep,pV[3]-ysep)[2],pV[3],pV[4]+\
        kickf(pV[1]-xsep,pV[3]-ysep)[1],l,dp>;# subtracts transforms to beam c/s
        return Record('l'=len,'k1'=k,'kind'=cat('Wire'),R=thismodule:-Quad(len,k*len,1):-R,\
                      'Eref'=Ener,'TF'=trackFunction);

    end proc; # Wire

#---------Foil--------------------------------------
    (* Foil(len,X0,Eref:=0)
   Implements a foil crossed by the beam; e.g. a window
   Needs the beam energy to work.
   Eloss not yet modeled => thicker foils are inaccurate (even if sliced!)
    len in m; X0 in m; Ener in MeV
 *)
    Foil:=proc(len,X0,Ener:=0)
    local EnerRef,trackFunction,Distribution;
    local rmsAngleTrans,Trans,Azimuth,RandFill,RandIndex;
    local RandAngle,RandAzimuth;

        EnerRef:=`if`(Ener<>0,Ener,Edefault);

        if (EnerRef=0) then
            error "Energy is 0. Cannot do scattering";
            return;
        end if;

        rmsAngleTrans:=MSTh0(X0,len,Ener);
        print("Rms scattering angle is ",rmsAngleTrans);
        Trans:=Statistics:-RandomVariable(Normal(0,rmsAngleTrans));
        Azimuth:=Statistics:-RandomVariable(Uniform(-Pi,Pi));
        RandIndex:=-1; # BeamParticles is not yet known

        RandFill:=proc()
            RandAngle:=Statistics:-Sample(Trans,BeamParticles);
            RandAzimuth:=Statistics:-Sample(Azimuth,BeamParticles);
            RandIndex:=0;
        end proc; # RandFill

        trackFunction:=proc(cV)
        local dxp,dyp;
            if (RandIndex >= BeamParticles or RandIndex=-1) then RandFill() end if;
            RandIndex:=RandIndex+1;
            #print("In Foil trackFunction. RandIndex =",RandIndex);
            dxp:=RandAngle[RandIndex]*cos(RandAzimuth[RandIndex]);
            dyp:=RandAngle[RandIndex]*sin(RandAzimuth[RandIndex]);
            return cV+~<0,dxp,0,dyp,0,0>;
        end proc; # trackFunction

        return Record('l'=len,'kind'=cat('Foil'),'R'=LinearAlgebra:-IdentityMatrix(6),\
                      'Eref'=Ener,'dE'=0,'TF'=trackFunction);

    end proc; # Foil

#------------ Manipulators (e.g. putting lines together) -------------

# Overloading the "." operator to concatenate elements & trace beam ellipses
    `.`:=overload([
        proc(elements::seq(Element),beam::Beam:=NULL,$) option overload;
        local cumR:=LinearAlgebra:-IdentityMatrix(6),cuml:=0,element;
        local lSigma:=Matrix(6,6);

            for element in [elements] do
                cumR:=LinearAlgebra:-Multiply(cumR,element:-R):
                cuml:=cuml+element:-l:
            end do;

            if (beam<>NULL) then # we trace an ellipse; generate a Beam
                lSigma:=LinearAlgebra:-Multiply(cumR,LinearAlgebra:-Multiply(beam:-Sigma,LinearAlgebra:-Transpose(cumR)));
                return Record(Particle=beam:-Particle,Energy=beam:-Energy,Mass=beam:-Mass,\
                              Charge=beam:-Charge,G=beam:-G,Gammar=beam:-Gammar,Sigma=lSigma);
            else # This is a beam line
                return Record(l=cuml,R=cumR);
            end if;

        end proc, # concatenation of a Lattice element

        proc(elements::(algebraic)) option overload;
            return :-`.`(elements)
        end proc # anything else
                  ]); # End of the overload

# Overloading the "^" operator for multiple instances of the same Element
    `^`:=overload([
        proc(element::Element,n::algebraic) option overload;
        description "The ^ operator concatenates n instances of element.";
        local i;
            return Record('l'=element:-l*n,\
                          `if`(member(a,element),'a'=element:-a*n,NULL),\
                          `if`(member(rho,element),'rho'=element:-rho,NULL),\
                          `if`(member(kind,element),'kind'=element:-kind,NULL),\
                          'dE'=dE*n,'R'=LinearAlgebra:-MatrixPower(element:-R,n),\
                          `if`(type(n,posint),'BL'=[seq(element,i=1..n)],NULL),\
                          `TF`=element:-TF@@n);
        end proc, # Power of a Lattice Element
        proc(e::algebraic,n::algebraic) option overload;
            return :-`^`(e,n);
        end proc # anything else
                  ]); # end of the overload

# The DefineLine function also stores element names & makes a BeamLine:
    DefineLine:=proc(Components::seq(Element),$)
    local i,component;
    local cumTF:=(X) -> X;
    local cumR:=LinearAlgebra:-IdentityMatrix(6),cuml:=0,cumdE:=0,cumang:=0;
    local Line:=[Components];

        for component in [Components] do
            cumR:=LinearAlgebra:-Multiply(component:-R,cumR):
            cuml:=cuml+component:-l:
            cumdE:=cumdE+component:-dE:
            cumTF:=component:-TF@eval(cumTF);

            if ((component:-kind=cat(Bend)) or (component:-kind=cat(BeamLine))) then
                cumang:=cumang+component:-a
            end if:

        end do;

        return Record('l'=cuml,'a'=cumang,'R'=cumR,'dE'=cumdE,'kind'=cat('BeamLine'),\
                      'TF'=cumTF,'BL'=Line);
    end proc; # DefineLine

# ExpandLine creates instances of each element thus allowing things like energy sawtooth
    ExpandLine:=proc(Components::seq(anything),$)
    local i:=0;# i always points to the last entry; i.e. need increment before using.
    local component,line;
        unprotect('Elems');
        line:=[Components];
        Elems:=Vector(1,0,datatype=anything);

        if (type(line[1],'Element')) then
            if (member(Eref,line[1])) then Ea:=line[1]:-Eref end if;
        else # has to be a line from DefineLine2
            Ea:=line[1][1]:-Eref;
        end if; # Ea now has the initial energy

        if (Ea=0) then Ea:=Edefault end if; # if we have not set Eref, use the default.

        sAct:=0; # init s coordinate
        recursions:=1;
        index:=0; # point to location before actual one
        expand_line(line); # fills the Elems Vector
        if (recursions <> 0) then print(recursions) end if;
        return copy(Elems);

    end proc; # ExpandLine

    expand_line:=proc(line::{list,Vector}) # here line is a list or a Vector
    local idx; # this is the local index into the (sub-)line
    local ii;

        for ii from 1 to numelems(line) do      # loop over line

            if (type(line[ii],'ExpandedLine')) then # this component is an expanded subline, include it
                recursions:=recursions+1;
                expand_line(line[ii])              # expand subline. Updates index
            elif (type(line[ii],'BeamLine')) then # this component is a BeamLine
                recursions:=recursions+1;
                expand_line(line[ii]:-BL); # expand it
            else                           # this is an element
                index:=index+1;

                if (type(line[ii]:-dE,procedure)) then
                    Ea:=Ea+line[ii]:-dE(Ea)
                else
                    Ea:=Ea+line[ii]:-dE
                end if;

                if (indets(Ea)={}) then Ea:=evalf(Ea) end if; # if symbolic, keep symbolic
                sAct:=sAct+line[ii]:-l;

                if (member('Eact',line[ii]) and has(line[ii],'s')) then # for some reason member does not work with s
                    Elems(index):=Record[line[ii],packed]();
                    Elems[index]:-Eact:=Ea;
                    Elems[index]:-s:=sAct
                elif (member('Eact',line[ii]) and not has(line[ii],'s')) then
                    Elems(index):=Record[line[ii],packed]('s'=sAct);
                    Elems[index]:-Eact:=Ea;
                elif (not member('Eact',line[ii]) and has(line[ii],'s')) then
                    Elems(index):=Record[line[ii],packed]('Eact'=Ea);
                    Elems[index]:-s:=sAct
                else
                    Elems(index):=Record[line[ii],packed]('Eact'=Ea,'s'=sAct)
                end if;

            end if; # line[ii] is a Vector

        end do; # loop over line

        recursions:=recursions-1; # keep track of rec. levels

        return;

    end proc; # expand_line

# LumpLine lumps an expandeed line into a BeamLine
    LumpLine:=proc(Eline::ExpandedLine)
    local i;
    local cumTF:=(X) -> X;
    local cumR:=LinearAlgebra:-IdentityMatrix(6),cuml:=0,cumdE:=0,cumang:=0;
    local Line:=[Eline];

        for i from 1 to numelems(Eline) do
            cuml:=cuml+Eline[i]:-l;
            cumR:=LinearAlgebra:-Multiply(Eline[i]:-R,cumR);
            cumdE:=cumdE+Eline[i]:-dE;
            cumTF:=Eline[i]:-TF@eval(cumTF);

            if ((Eline[i]:-kind=cat(Bend)) or (Eline[i]:-kind=cat(BeamLine))) then
                cumang:=cumang+Eline[i]:-a
            end if:

        end do; # i loop over elements

        return Record('l'=cuml,'a'=cumang,'R'=cumR,'dE'=cumdE,'kind'=cat('BeamLine'),\
                      'TF'=cumTF,'BL'=Line);

    end proc; # LumpLine

# The DefineMachine function stores machine names but also a comment (for now):
    DefineMachine:=proc(title::string,Components::seq(Element),$)
    local i,component;
    local cumR:=LinearAlgebra:-IdentityMatrix(6),cuml:=0,cumdE:=0;
    local Line:=[Components];

        for component in [Components] do
            cumR:=LinearAlgebra:-Multiply(cumR,component:-R):
            cuml:=cuml+component:-l:
            cumdE:=cumdE+component:-dE:
        end do;

        return Record(l=cuml,R=cumR,Title=title,kind=cat('Machine'),BL=Line);
    end proc;

# DefineBeam defines beam-particles and a beam energy. It will be used to facilitate s.r. and calculations. beam emittance will also go here.
    DefineBeam:=proc(particle::name,Ebeam,{Sigmaa:=0},epsx:=0,epsy:=0,twss:=0::Vector,dE_E:=0,disp:=0::Vector,P:=0::Vector,Npart:=0.) uses Statistics;
    description "Defines a beam. Emittances optional, maybe calculated if possible. If twiss and disp given, sets up ∑ matrix. If Npart ≠ 0, generates a Gaussian beam";
    local Sigmal:=Matrix(6,6,fill=0);
    local psi1:=Vector(2);
    local f,i;
    local rootepsx:=sqrt(epsx);
    local rootbetax:=sqrt(twss[1]);
    local rootepsy:=sqrt(epsy);
    local rootbetay:=sqrt(twss[4]);
    local PartVector,Dist,AmpDist,AngDist,Ampx,Angx,Ampy,Angy,qx,px,qy,py,dEE;
    local detSigx,detSigy,detSigp,wx,wxp,wy,wyp,wl,wdp,wxxp,wyyp,wxxpx,wxxpxp,wyypy,wyypyp;

        BeamEnergy:=Ebeam; # [MeV]
        BeamParticleRestMass:=ParticleTable[particle][1]; # [MeV]
        BeamGammar:=(BeamEnergy+BeamParticleRestMass)/BeamParticleRestMass;
        BeamBetar:=sqrt(1-1/BeamGammar^2);

        if (Sigmaa<>0) then
            Sigmal:=copy(Sigmaa);
        elif type(twss,Vector) then
            Sigmal[1,1]:=twss[1]*epsx;
            Sigmal[2,2]:=twss[3]*epsx;
            Sigmal[1,2]:=-twss[2]*epsx;
            Sigmal[2,1]:=Sigmal[1,2];
            Sigmal[3,3]:=twss[4]*epsy;
            Sigmal[4,4]:=twss[6]*epsy;
            Sigmal[3,4]:=-twss[5]*epsy;
            Sigmal[4,3]:=Sigmal[3,4];
            Sigmal[5,5]:=1;
            Sigmal[6,6]:=(dE_E)^2;
            Sigmal[1,6]:=disp[1]*Sigmal[6,6];
            Sigmal[6,1]:=Sigmal[1,6];
            Sigmal[2,6]:=disp[2]*Sigmal[6,6];
            Sigmal[6,2]:=Sigmal[2,6];
        end if;

        detSigx:=LinearAlgebra:-Determinant(LinearAlgebra:-SubMatrix(Sigmal,1..2,1..2));
        detSigy:=LinearAlgebra:-Determinant(LinearAlgebra:-SubMatrix(Sigmal,3..4,3..4));
        detSigp:=LinearAlgebra:-Determinant(LinearAlgebra:-SubMatrix(Sigmal,5..6,5..6));
        print("Determinants",detSigx,detSigy,detSigp);
        if (detSigp<0.) then detSigp:=0 end if; # kludge to track down Maple 2015 issue

        if (Sigmal[1,1]>0) then
            wx:=sqrt(Sigmal[1,1]);
            wxp:=sqrt(Sigmal[2,2]-Sigmal[1,2]*Sigmal[1,2]/Sigmal[1,1]);
            wxxp:=sign(Sigmal[1,2])*sqrt(Sigmal[1,2]*Sigmal[1,2]/Sigmal[1,1]);
        else
            wx:=0;
            wxp:=0;
            wxxp:=0;
        end if;

        if (Sigmal[3,3]>0) then
            wy:=sqrt(Sigmal[3,3]);
            wyp:=sqrt(Sigmal[4,4]-Sigmal[3,4]*Sigmal[3,4]/Sigmal[3,3]);
            wyyp:=sign(Sigmal[3,4])*sqrt(Sigmal[3,4]*Sigmal[3,4]/Sigmal[3,3]);
        else
            wy:=0;
            wyp:=0;
            wyyp:=0;
        end if;

#        wx:=sqrt(Sigmal[1,1]);
#        wxp:=sqrt(Sigmal[2,2]-Sigmal[1,2]*Sigmal[1,2]/Sigmal[1,1]);
#        wy:=sqrt(Sigmal[3,3]);
#        wyp:=sqrt(Sigmal[4,4]-Sigmal[3,4]*Sigmal[3,4]/Sigmal[3,3]);
#        wxxp:=sign(Sigmal[1,2])*sqrt(Sigmal[1,2]*Sigmal[1,2]/Sigmal[1,1]);
#        wyyp:=sign(Sigmal[3,4])*sqrt(Sigmal[3,4]*Sigmal[3,4]/Sigmal[3,3]);

        if type(P,Vector) then
            f:=sqrt(2*(P[3]+1));
            psi1[1]:=(1+P[3])/f + I*0;
            psi1[2]:=P[1]/f + I*P[2]/f;
        end if;

        if (Npart>0. and type(twss,Vector)) then
            PartVector:=table();
            # We generate Gaussian distributions with sigma=1. These
            # get transformed below using the Sigma matrix

            AmpDist:=RandomVariable(Distribution(PDF='t -> piecewise(t>-1E-6,exp(-t^2)*2*t,0.)'));
            AngDist:=RandomVariable(Uniform(-Pi,Pi));
            Ampx:=Sample(AmpDist,Npart,method=[envelope,range=0..10]);
            Angx:=Sample(AngDist,Npart);
            Ampy:=Sample(AmpDist,Npart,method=[envelope,range=0..10]);
            Angy:=Sample(AngDist,Npart);
            qx:=(Ampx)*~sin~(Angx);
            px:=(Ampx)*~cos~(Angx);
            qy:=(Ampy)*~sin~(Angy);
            py:=(Ampy)*~cos~(Angy);
            dEE:=Sample(AmpDist,Npart,method=[envelope,range=0..10]);
            seq(assign('PartVector'[i]=evalf~(Vector[column](<qx[i]*wx + 0*px[i]*wxxp,\
                                                             qx[i]*wxxp + px[i]*wxp,\
                                                             qy[i]*wy + 0*py[i]*wyyp,\
                                                             qy[i]*wyyp + py[i]*wyp,\
                                                             0,dEE[i]*dE_E>))),i=1..Npart);
        end if;

        return Record('Particle'=particle,'Energy'=Ebeam,'Mass'=BeamParticleRestMass,'Charge'=ParticleTable[particle][2],'G'=ParticleTable[particle][3],'Gammar'=BeamGammar,'Betar'=BeamBetar,'Sigma'=Sigmal,'psi'=psi1,'Nparticles'=Npart,'Coordinates'=PartVector);
    end proc; # DefineBeam

#------------ Manipulations of beam lines or elements

# Subs is a substitute command that works on Lattice Elements.
# It accepts a sequence of replacement equations, but not a list.
    Subs:=proc(eqn::seq(equation),elemt::Element,$) option remember;
    description "subs command for Elements or BeamLines.";
    local qs:=[exports(elemt)];
    local i,j;
    local selemt:=[seq(qs[i],i=1..numelems(qs))];
    local mysubs;

        mysubs:=proc(eqn::seq(equation),var,xpr)
        local rt;

            if (var = 'TF') then # this is the track function, needs special treatment
                return unapply(subs(eqn,xpr(<pV[1],pV[2],pV[3],pV[4],pV[5],pV[6]>)),pV)
            else
                return subs(eqn,xpr);
            end if;

        end proc; # mysubs

        if (type(elemt,BeamLine)) then
            return Record(seq(`if`(qs[i]<>BL,qs[i]=mysubs(eqn,qs[i],elemt[qs[i]]),qs[i]=[seq((Subs~(eqn,elemt[qs[i]][j])),j=1..numelems(elemt[qs[i]]))]),i=1..numelems(qs)));
        else
            return Record(seq(qs[i]=mysubs(eqn,qs[i],elemt[qs[i]]),i=1..numelems(qs)));
        end if;

    end proc; # Subs

    SRotate:=proc(elemt::Element,angle)
    description "Rotate an Element about s by angle";
    local C:=cos(angle);
    local S:=sin(angle);
    local ekind;
    local Rmat;
    local Mr:=Matrix(6,6,{(1,1)=C,(1,3)=S,\
                          (2,2)=C,(2,4)=S,\
                          (3,1)=-S,(3,3)=C,\
                          (4,2)=-S,(4,4)=C,\
                          (5,5)=1,(6,6)=1},\
                     fill=0);
    local Mri:=Matrix(6,6,{(1,1)=C,(1,3)=-S,\
                           (2,2)=C,(2,4)=-S,\
                           (3,1)=S,(3,3)=C,\
                           (4,2)=S,(4,4)=C,\
                           (5,5)=1,(6,6)=1},\
                      fill=0);
    local Mrot:=LinearAlgebra:-Multiply(Mri,LinearAlgebra:-Multiply(elemt:-R,Mr));

        if (type(elemt,'BeamLine')) then
            error "Can't do beam lines or machines yet";
        end if;

        ekind:=elemt:-kind;
        print(ekind);
        Rmat:=Record[elemt]();
        Rmat:-R:=Mrot;
        return Rmat;
    end proc; # SRotate

#------------ Plot the beam ellipses ---------------

    BeamPlot:=proc(beam::Beam)
    description "Returns PLOT Array with the phase space ellipses if ∑ matrix exists";
    local Mx:=evalf~(LinearAlgebra:-MatrixInverse(LinearAlgebra:-SubMatrix(beam:-Sigma,1..2,1..2)));
    local My:=evalf~(LinearAlgebra:-MatrixInverse(LinearAlgebra:-SubMatrix(beam:-Sigma,3..4,3..4)));
    local x,xp,xp0,y,yp,yp0,l,dp,sp0;
    local X:=<x,xp>,Y:=<y,yp>,S:=<l,dp>;
    local ellipsx,ellipsy,ellipss,xmaxEll,ymaxEll;
    local xmax,ymax,xmin,ymin,dpmax,xavg,xpavg,yavg,ypavg,dpavg,xvec,xpvec,yvec,ypvec,lvec,dpvec,ii;
    local PlArray;

        if (beam:-Sigma=NULL) then
            print("No Sigma matrix in beam");
            return;
        end if;

        ellipsx:=1=evalf(LinearAlgebra:-Multiply(LinearAlgebra:-Transpose(X),LinearAlgebra:-Multiply(Mx,X)));
        ellipsy:=1=evalf(LinearAlgebra:-Multiply(LinearAlgebra:-Transpose(Y),LinearAlgebra:-Multiply(My,Y)));
        xp0:=solve(ellipsx,xp);
        yp0:=solve(ellipsy,yp);
        xmaxEll:=evalf(sqrt(beam:-Sigma[1,1]));
        ymaxEll:=evalf(sqrt(beam:-Sigma[3,3]));
        dpmax:=0; # evalf(sqrt(beam:-Sigma[5,5]));

        if (beam:-Nparticles = 0.) then
            PlArray:=Array(1..2);
            PlArray[1]:=plot([xp0[1],xp0[2]],x=min(-xmaxEll,-ymaxEll)*1.2..max(xmaxEll,ymaxEll)*1.2,\
                             color=red,labels=[x (m),xp (r)]);
            PlArray[2]:=plot([yp0[1],yp0[2]],y=min(-xmaxEll,-ymaxEll)*1.2..max(xmaxEll,ymaxEll)*1.2,\
                             color="Green",labels=[y (m),yp (r)]);

            if (dpmax >0) then #Note prog indexing to extend PlArray!
                PlArray(3):=plot([sp0[1],sp0[2]],dp=-dpmax..dpmax,color=blue);
            end if;

        else
            PlArray:=Array(1..2,1..2);
            xvec:=[seq(evalf(beam:-Coordinates[ii][1]),ii in indices(beam:-Coordinates,'nolist'))];
            xpvec:=[seq(evalf(beam:-Coordinates[ii][2]),ii in indices(beam:-Coordinates,'nolist'))];
            yvec:=[seq(evalf(beam:-Coordinates[ii][3]),ii in indices(beam:-Coordinates,'nolist'))];
            ypvec:=[seq(evalf(beam:-Coordinates[ii][4]),ii in indices(beam:-Coordinates,'nolist'))];
            xavg:=Statistics:-Mean(xvec);
            xpavg:=Statistics:-Mean(xpvec);
            yavg:=Statistics:-Mean(yvec);
            ypavg:=Statistics:-Mean(ypvec);
            xmin:=Re(min(xavg-xmaxEll,min(xvec)));
            xmax:=Re(max(xavg+xmaxEll,max(xvec)));
            ymin:=Re(min(yavg-ymaxEll,min(yvec)));
            ymax:=Re(max(yavg+ymaxEll,max(yvec)));

            PlArray[1,1]:=plots:-display(plots:-pointplot(xvec,xpvec,symbol=point,color="Red"),\
                                         plot([eval(xp0[1],x=x-xavg)+xpavg,eval(xp0[2],x=x-xavg)+xpavg],\
                                              x=xmin..xmax,color="Maroon"),labels=[typeset('x'," (m)"),typeset('xp'," (rad)")]);
            PlArray[1,2]:=plots:-display(plots:-pointplot(yvec,ypvec,symbol=point,color="Green"),\
                                         plot([eval(yp0[1],y=y-yavg)+ypavg,eval(yp0[2],y=y-yavg)+ypavg],\
                                              y=ymin..ymax,color="DarkGreen"),labels=[typeset('y'," (m)"),typeset('yp'," (rad)")]);
            PlArray[2,1]:=plots:-display(plots:-pointplot(xvec,yvec,symbol=point,color="Red"),\
                                         labels=[typeset('x'," (m)"),typeset('y'," (m)")],scaling=constrained);
            PlArray[2,2]:=plots:-display(plots:-pointplot(xpvec,ypvec,symbol=point,color="Green"),\
                                         labels=[typeset('xp'," (rad)"),typeset('yp'," (rad)")],scaling=constrained);

            if (dpmax >0) then #Note prog indexing to extend PlArray!
                PlArray(3):=plot([sp0[1],sp0[2]],dp=-dpmax..dpmax,color=blue);
            end if;

        end if; # Npart>0, i.e. particles to be plotted.

        return PlArray;
    end proc; # BeamPlot

#------------ Output for lattice programs

# The MAD8 output command accepts only a machine
    Mad8Form:=proc(mach::Machine,parmlist::uneval)
    description "Mad8Form(mach,parmlist) provides Mad8 elements, beamlines and parameters.";
    local OutString::string;
    local thinquad:=false;
    local elem,machnam,parm,sortedIndices;
        machnam:=convert('mach',string);
        unprotect(DriftTable);unprotect(QuadTable);unprotect(LineTable);
        DriftTable:=table();QuadTable:=table();BendTable:=table();LineTable:=table();
        OutTables(mach); # This builds all the tables
        OutString:=cat("Title, \"",mach:-Title,"\"\n");# print(OutString);

# Output the Bends
        sortedIndices:=sort([indices(BendTable, 'nolist')]);

        for elem in sortedIndices do
            OutString:=cat(OutString,elem,":  SBEND,  L=",convert(BendTable[elem]:-l,string),", angle=",convert(BendTable[elem]:-a,string),", k1=",convert(BendTable[elem]:-n,string),"\n");
        end do;

# Output the Quads
        sortedIndices:=sort([indices(QuadTable, 'nolist')]);

        for elem in sortedIndices do
            if (eval(QuadTable[elem]:-l,1)=0) then
                thinquad:=true;
                OutString:=cat(OutString,elem,":  QUAD,  L=lqzero",", k1=",convert(eval(QuadTable[elem]:-k1l,1),string),\
                               "/lqzero !! Zero-length quad !! \n")
            else
                OutString:=cat(OutString,elem,":  QUAD,  L=",convert(eval(QuadTable[elem]:-l,1),string),", k1=",\
                               convert(eval(QuadTable[elem]:-k1l/QuadTable[elem]:-l,1),string),"\n");
            end if;
        end do;

        OutString:=cat(OutString,"\n"); # blank line for readibility

# Output the Drifts
        sortedIndices:=sort([indices(DriftTable, 'nolist')]);

        for elem in sortedIndices do
            OutString:=cat(OutString,elem,":  DRIFT,  L=",convert(eval(DriftTable[elem]:-l,1),string),"\n");
        end do;

        OutString:=cat(OutString,"\n"); # blank line for readibility

# Output the BeamLines
        for elem in indices(LineTable) do

            if (SearchText("Record(",elem[],1..7)<>1) then # avoid "Record(…)" as beam-line name for unassigned machine.
                OutString:=cat(OutString,elem[],":  Line=(",substring(convert(eval(LineTable[elem[]],1),string),2..-2),")\n");
            end if;

        end do;

        OutString:=cat(OutString,"\n"); # blank line for readibility

# Output the parameters
        if (type(parmlist,list)) then

            for parm from 1 to nops(parmlist) do

                if (type(parmlist[parm],'equation')) then
                    OutString:=cat(OutString,convert(lhs(parmlist[parm]),string)," := ",convert(eval(rhs(parmlist[parm])),string),"\n")
                else
                    OutString:=cat(OutString,convert(parmlist[parm],string)," := ",convert(eval(parmlist[parm]),string),"\n")
                end if;

            end do;

        else error "parmlist is not a list"
        end if;

        if(thinquad) then OutString:=cat(OutString,"lqzero:=1E-04 !! For zero-length quads !! \n") end if;

        return OutString;
    end proc; # Mad8Form

# The OutTables aux command accepts Elements so we can recurse...
    OutTables:=proc(mach::seq(Element),$)
    local i,j,k,line,component,machine,machnam;
        machnam:=convert('mach',string);
        machine:=mach; # prevent this from being overridden

# need to find any beamlines
        if (type(machine,BeamLine)) then line:=machine:-BL; # this is the beam line
            LineTable[machnam]:=line;
        else line:=[machine]; # must be an element
        end if;

        if (recursions>=10) then return end if; #Emergency stop for debugging

        for component in line do
            if (component:-kind=cat(Drift)) then
                Driftidx:=Driftidx+1;
                DriftTable[component]:=component;
            elif (component:-kind=cat(Bend)) then
                Bendidx:=Bendidx+1;
                BendTable[component]:=component;
            elif (component:-kind=cat(Quad)) then
                Quadidx:=Quadidx+1;
                QuadTable[component]:=component;
            else
                recursions:=recursions+1;
                OutTables(component);
            end if;

        end do;

    end proc;# Outtables

#----------------- Prepare Vectors for graphic output ---------------
    LatticePlot:=proc(bl::anything,Twiss::Vector,Disp::Vector)
    description "Returns PLOT Array with lattice functions by element";
    local i,j,k,nelems,line,component,blnam;
    local PlArray:=Array(1..2);
        blnam:=convert('bl',string);
        i:=1;j:=1;k:=1;

        betaxT:=Vector(1,0,datatype=float[8]);
        alphaxT:=Vector(1,0,datatype=float[8]);
        etaxT:=Vector(1,0,datatype=float[8]);
        etapxT:=Vector(1,0,datatype=float[8]);
        betayT:=Vector(1,0,datatype=float[8]);
        alphayT:=Vector(1,0,datatype=float[8]);
        etayT:=Vector(1,0,datatype=float[8]);
        etapyT:=Vector(1,0,datatype=float[8]);

        ST(i):=0;
        dET(i):=0;
        ElnamT(i):="Initial";
        TwissT:=Twiss;
        EtaT:=Disp;
        betaxT(i):=evalf(Twiss[1]);
        alphaxT(i):=evalf(Twiss[2]);
        etaxT(i):=evalf(Disp[1]);
        etapxT(i):=evalf(Disp[2]);
        betayT(i):=evalf(Twiss[4]);
        alphayT(i):=evalf(Twiss[5]);
        etayT(i):=evalf(Disp[3]);
        etapyT(i):=evalf(Disp[4]);

        if (type(bl,'BeamLine')) then
            i:=i+1;
            nelems:=lattab(bl,i); # this builds the tables
        elif (type(bl,'ExpandedLine')) then
            i:=i+1;
            nelems:=lattab(bl,i); # lattab works on Vectors also
        else
            error "Wrong type for beam line: %1",bl;
            return;
        end if;

        PlArray:=[plot(Lattice:-ST,Lattice:-betaxT),plot(Lattice:-ST,Lattice:-betayT,color=green),\
                  plot(ST,etaxT,color=blue)];

        return PlArray;

    end proc; # LatticePlot

### Subproc lattab
    lattab:=proc(blin::anything,idx::integer)# option trace;
    local line,component,idr,nelems;

        if type(blin,'BeamLine') then
            line:=blin:-BL; # beamline (list of lines or Elements)
            nelems:=nops(line);
        elif type(blin,'ExpandedLine') then # ExpandedLine
            line:=blin;
            nelems:=numelems(line);
        elif type(blin,'Element') then
            line:=[blin]; # must be an element
            nelems:=1;
        else
            error "Unrecognized element: $1",blin;
            return;
        end if;

# pre-allocate tables
        ST(idx+nelems-1):=0;
        betaxT(idx+nelems-1):=0;
        alphaxT(idx+nelems-1):=0;
        etaxT(idx+nelems-1):=0;
        etapxT(idx+nelems-1):=0;
        betayT(idx+nelems-1):=0;
        alphayT(idx+nelems-1):=0;
        etayT(idx+nelems-1):=0;
        etapyT(idx+nelems-1):=0;
        ElnamT(idx+nelems-1):="";
        dET(idx+nelems-1):=0;
        idr:=idx;

        for component in line do

            if type(component,BeamLine) then # this is another beam line
                recursions:=recursions+1;
                idr:=lattab(component,idr);
                recursions:=recursions-1;
            else
                ST(idr):=ST(idr-1)+component:-l; # running total of length
                ElnamT(idr):=convert(component,string);
                TwissT:=evalf~(TwissTran(component,TwissT)); # propagate
                EtaT:=evalf~(EtaTran(component,EtaT)); # propagate
                betaxT(idr):=TwissT[1];
                alphaxT(idr):=TwissT[2];
                etaxT(idr):=EtaT[1];
                etapxT(idr):=EtaT[2];
                betayT(idr):=TwissT[4];
                alphayT(idr):=TwissT[5];
                etayT(idr):=EtaT[3];
                etapyT(idr):=EtaT[4];

                if (member('dE',component)) then

                    if type(component:-dE,'procedure') then
                        dET(idr):=evalf(dET(idr-1)+component:-dE(Edefault))
                    else
                        dET(idr):=evalf(dET(idr-1)+component:-dE)
                    end if;

                else
                    dET(idr):=evalf(dET(idr-1))
                end if;

                idr:=idr+1;
            end if

        end do;

        return idr; # hand back the index
    end proc; # lattab
#------------------- End LatticePlot

#------------------- definitions of the basic lattice-function projectors
    sinmux:=(line::{Element,ExpandedLine,Matrix})->sqrt(1-cosmux(line)^2);
    sinmuy:=(line::{Element,ExpandedLine,Matrix})->sqrt(1-cosmuy(line)^2);
    twissx:=(line::{Element,ExpandedLine,Matrix})-><(betax(line)),alphax(line),(gammax(line))>;
    twissy:=(line::{Element,ExpandedLine,Matrix})-><(betay(line)),alphay(line),(gammay(line))>;
    twiss:=(line::{Element,ExpandedLine,Matrix})-><twissx(line),twissy(line)>[..,1];
    dispersion:=(line::{Element,ExpandedLine,Matrix})->convert(<etax(line),etapx(line),etay(line),etapy(line)>,Vector);
    Hx:=(line::{Element,ExpandedLine,Matrix}) -> betax(line)*etapx(line)^2+2*alphax(line)*etax(line)*etapx(line)+gammax(line)*etax(line)^2;
    Hy:=(line::{Element,ExpandedLine,Matrix}) -> betay(line)*etapy(line)^2+2*alphay(line)*etay(line)*etapy(line)+gammay(line)*etay(line)^2;
    H:= (line::{Element,ExpandedLine,Matrix}) -> <Hx(line),Hy(line)>;

#------------------- definitions of the basic lattice-function projectors to include ExpandedLine
    cosmux:=proc(line::{Element,ExpandedLine,Matrix})
    local i,R;
        if type(line,'Element') then
            return 1/2*(line:-R[1,1]+line:-R[2,2]);
        elif type(line,'Matrix') then
            return 1/2*(line[1,1]+line[2,2]);
        elif type(line,'ExpandedLine') then
            R:=copy(line[1]:-R);
            for i from 2 to numelems(line) do
                R:=LinearAlgebra:-MatrixMatrixMultiply(line[i]:-R,R)
            end do;

            return 1/2*(R[1,1]+R[2,2]);
        else
            error "Unknown type of argument. Only Element, BeamLine and ExpandedLine are supported";
        end if;
    end proc; # cosmux

    cosmuy:=proc(line::{Element,ExpandedLine,Matrix})
    local i,R;
        if type(line,'Element') then
            return 1/2*(line:-R[3,3]+line:-R[4,4]);
        elif type(line,'Matrix') then
            return 1/2*(line[3,3]+line[4,4]);
        elif type(line,'ExpandedLine') then
            R:=copy(line[1]:-R);

            for i from 2 to numelems(line) do
                R:=LinearAlgebra:-MatrixMatrixMultiply(line[i]:-R,R)
            end do;

            return 1/2*(R[3,3]+R[4,4]);
        else
            error "Unknown type of argument. Only Element, BeamLine and ExpandedLine are supported";
        end if;
    end proc; # cosmuy

    nux:=proc(line::{Element,ExpandedLine,Matrix},tw::Vector:=NULL)
    local i,R,nu,twis;
        if (tw<>NULL) then # we have a twiss vector so use that

            if type(line,'Element') then
                return arctan(line:-R[1,2]/(tw[1]*line:-R[1,1]-tw[2]*line:-R[1,2]))/(2*Pi);
            elif type(line,'Matrix') then
                return arctan(line[1,2]/(tw[1]*line[1,1]-tw[2]*line[1,2]))/(2*Pi);
            elif type(line,'ExpandedLine') then # ExpandedLine
                nu:=0;
                twis:=copy(tw);

                for i from 1 to numelems(line) do
                    nu:=nu+arctan(line[i]:-R[1,2]/(twis[1]*line[i]:-R[1,1]-twis[2]*line[i]:-R[1,2]))/(2*Pi);
                    twis:=TwissTran(line[i],twis);
                end do; # i

                return nu;
            end if; # Expanded line or element

        else # no twiss vector, use matched solution

            if (type(line,'Element') or type(line,'Matrix')) then
                return arccos(cosmux(line))/(2*Pi);
            elif type(line,'ExpandedLine') then # ExpandedLine
                R:=copy(line[1]:-R);

                for i from 2 to numelems(line) do
                    R:=LinearAlgebra:-MatrixMatrixMultiply(line[i]:-R,R)
                end do; # i

                return arccos(1/2*(R[1,1]+R[2,2]))/(2*Pi);
            end if; # expanded line or element

        end if; # twiss vector goven or not

    end proc; # nux

    nuy:=proc(line::{Element,ExpandedLine,Matrix},tw::Vector:=NULL)
    local i,R,nu,twis;

        if (tw<>NULL) then # we have a twiss vector so use that

            if type(line,'Element') then
                return arctan(line:-R[3,4]/(tw[4]*line:-R[3,3]-tw[5]*line:-R[3,4]))/(2*Pi);
            elif type(line,'Matrix') then
                return arctan(line[3,4]/(tw[4]*line[3,3]-tw[5]*line[3,4]))/(2*Pi);
            elif type(line,'ExpandedLine') then # ExpandedLine
                nu:=0;
                twis:=copy(tw);

                for i from 1 to numelems(line) do
                    nu:=nu+arctan(line[i]:-R[3,4]/(twis[4]*line[i]:-R[3,3]-twis[5]*line[i]:-R[3,4]))/(2*Pi);
                    twis:=TwissTran(line[i],twis);
                end do; # i

                return nu;
            end if; # Expanded line or element

        else # no twiss vector, use matched solution

            if (type(line,'Element') or type(line,'Matrix')) then
                return arccos(cosmuy(line))/(2*Pi);
            elif type(line,'ExpandedLine') then
                R:=copy(line[1]:-R);

                for i from 2 to numelems(line) do
                    R:=LinearAlgebra:-MatrixMatrixMultiply(line[i]:-R,R)
                end do;

                return arccos(1/2*(R[3,3]+R[4,4]))/(2*Pi);
            end if; # ExpandedLine or Element

        end if; # twiss vector given or not

    end proc; # nuy

    betax:=proc(line::{Element,ExpandedLine,Matrix})
    local i,R,bx,bxnum;
        if type(line,'Element') then
            bx:=line:-R[1,2]/sinmux(line)
        elif type(line,'Matrix') then
            bx:=line[1,2]/sinmux(line)
        elif type(line,'ExpandedLine') then
            R:=copy(line[1]:-R);

            for i from 2 to numelems(line) do
                R:=LinearAlgebra:-MatrixMatrixMultiply(line[i]:-R,R)
            end do;

            bx:=R[1,2]/sinmux(line);
        end if;

        bxnum:=evalf(bx);

        if (type(bxnum,numeric)) then
            if (abs(Im(bxnum))<1E-14 and Re(bxnum)<0) then bx:=-bx end if;
        end if;

        return bx;

    end proc; # betax

    alphax:=proc(line::{Element,ExpandedLine,Matrix})
    local i,R;
        if type(line,'Element') then
            return 1/2*(line:-R[1,1]-line:-R[2,2])/sinmux(line)
        elif type(line,'Matrix') then
            return 1/2*(line[1,1]-line[2,2])/sinmux(line)
        elif type(line,'ExpandedLine') then
            R:=copy(line[1]:-R);

            for i from 2 to numelems(line) do
                R:=LinearAlgebra:-MatrixMatrixMultiply(line[i]:-R,R)
            end do;

            return 1/2*(R[1,1]-R[2,2])/sinmux(line) ;
        end if;
    end proc; # alphax

    gammax:=proc(line::{Element,ExpandedLine,Matrix})
    local i,R,gx,gxnum;
        if type(line,'Element') then
            gx:=-line:-R[2,1]/sinmux(line)
        elif type(line,'Matrix') then
            gx:=-line[2,1]/sinmux(line)
        elif type(line,'ExpandedLine') then
            R:=copy(line[1]:-R);

            for i from 2 to numelems(line) do
                R:=LinearAlgebra:-MatrixMatrixMultiply(line[i]:-R,R)
            end do;

            gx:=-R[2,1]/sinmux(line) ;
        end if;

        gxnum:=evalf(gx);

        if (type(gxnum,numeric)) then
            if (abs(Im(gxnum))<1E-14 and Re(gxnum)<0) then gx:=-gx end if;
        end if;

        return gx;

    end proc; # gammax

    betay:=proc(line::{Element,ExpandedLine,Matrix})
    local i,R,bety,bynum;
        if type(line,'Element') then
            bety:=line:-R[3,4]/sinmuy(line)
        elif type(line,'Matrix') then
            bety:=line[3,4]/sinmuy(line)
        elif type(line,'ExpandedLine') then
            R:=copy(line[1]:-R);

            for i from 2 to numelems(line) do
                R:=LinearAlgebra:-MatrixMatrixMultiply(line[i]:-R,R)
            end do;

            bety:=R[3,4]/sinmuy(line);
        end if;

        bynum:=evalf(bety);

        if (type(bynum,numeric)) then
            if (abs(Im(bynum))<1E-14 and Re(bynum)<0) then bety:=-bety end if;
        end if;

        return bety;

    end proc; # betay

    alphay:=proc(line::{Element,ExpandedLine,Matrix})
    local i,R;
        if type(line,'Element') then
            return 1/2*(line:-R[3,3]-line:-R[4,4])/sinmuy(line)
        elif type(line,'Matrix') then
            return 1/2*(line[3,3]-line[4,4])/sinmuy(line)
        elif type(line,'ExpandedLine') then
            R:=copy(line[1]:-R);

            for i from 2 to numelems(line) do
                R:=LinearAlgebra:-MatrixMatrixMultiply(line[i]:-R,R)
            end do;

            return 1/2*(R[3,3]-R[4,4])/sinmuy(line) ;
        end if;
    end proc; # alphay

    gammay:=proc(line::{Element,ExpandedLine,Matrix})
    local i,R,gy,gynum;
        if type(line,'Element') then
            gy:=-line:-R[4,3]/sinmuy(line)
        elif type(line,'Matrix') then
            gy:=-line[4,3]/sinmuy(line)
        elif type(line,'ExpandedLine') then
            R:=copy(line[1]:-R);

            for i from 2 to numelems(line) do
                R:=LinearAlgebra:-MatrixMatrixMultiply(line[i]:-R,R)
            end do;

            gy:=-R[4,3]/sinmux(line) ;
        end if;

        gynum:=evalf(gy);

        if (type(gynum,numeric)) then
            if (abs(Im(gynum))<1E-14 and Re(gynum)<0) then gy:=-gy end if;
        end if;

        return gy;

    end proc; # gammay

    etax:=proc(line::{Element,ExpandedLine,Matrix})
    local i,R;
        if type(line,'Element') then
            return (line:-R[1,6]+line:-R[1,2]*line:-R[2,6]/(1-line:-R[2,2]))/(1-line:-R[1,1]-line:-R[2,1]*line:-R[1,2]/(1-line:-R[2,2]))
        elif type(line,'Matrix') then
            return (line[1,6]+line[1,2]*line[2,6]/(1-line[2,2]))/(1-line[1,1]-line[2,1]*line[1,2]/(1-line[2,2]))
        elif type(line,'ExpandedLine') then
            R:=copy(line[1]:-R);

            for i from 2 to numelems(line) do
                R:=LinearAlgebra:-MatrixMatrixMultiply(line[i]:-R,R)
            end do;

            return (R[1,6]+R[1,2]*R[2,6]/(1-R[2,2]))/(1-R[1,1]-R[2,1]*R[1,2]/(1-R[2,2]));
        end if;
    end proc; # etax

    etapx:=proc(line::{Element,ExpandedLine,Matrix})
    local i,R;
        if type(line,'Element') then
            return (line:-R[2,6]+line:-R[2,1]*line:-R[1,6]/(1-line:-R[1,1]))/(1-line:-R[2,2]-line:-R[2,1]*line:-R[1,2]/(1-line:-R[1,1]))
        elif type(line,'Matrix') then
            return (line[2,6]+line[2,1]*line[1,6]/(1-line[1,1]))/(1-line[2,2]-line[2,1]*line[1,2]/(1-line[1,1]))
        elif type(line,'ExpandedLine') then
            R:=copy(line[1]:-R);

            for i from 2 to numelems(line) do
                R:=LinearAlgebra:-MatrixMatrixMultiply(line[i]:-R,R)
            end do;

            return (R[2,6]+R[2,1]*R[1,6]/(1-R[1,1]))/(1-R[2,2]-R[2,1]*R[1,2]/(1-R[1,1]));
        end if;
    end proc; # etapx

    etay:=proc(line::{Element,ExpandedLine,Matrix})
    local i,R;
        if type(line,'Element') then
            return (line:-R[3,6]+line:-R[3,4]*line:-R[4,6]/(1-line:-R[4,4]))/(1-line:-R[3,3]-line:-R[4,3]*line:-R[3,4]/(1-line:-R[4,4]))
        elif type(line,'Matrix') then
            return (line[3,6]+line[3,4]*line[4,6]/(1-line[4,4]))/(1-line[3,3]-line[4,3]*line[3,4]/(1-line[4,4]))
        elif type(line,'ExpandedLine') then
            R:=copy(line[1]:-R);

            for i from 2 to numelems(line) do
                R:=LinearAlgebra:-MatrixMatrixMultiply(line[i]:-R,R)
            end do;

            return (R[3,6]+R[3,4]*R[4,6]/(1-R[4,4]))/(1-R[3,3]-R[4,3]*R[3,4]/(1-R[4,4]));
        end if;
    end proc; # etay

    etapy:=proc(line::{Element,ExpandedLine,Matrix})
    local i,R;
        if type(line,'Element') then
            return (line:-R[4,6]+line:-R[4,3]*line:-R[3,6]/(1-line:-R[3,3]))/(1-line:-R[4,4]-line:-R[4,3]*line:-R[3,4]/(1-line:-R[3,3]))
        elif type(line,'Matrix') then
            return (line[4,6]+line[4,3]*line[3,6]/(1-line[3,3]))/(1-line[4,4]-line[4,3]*line[3,4]/(1-line[3,3]))
        elif type(line,'ExpandedLine') then
            R:=copy(line[1]:-R);

            for i from 2 to numelems(line) do
                R:=LinearAlgebra:-MatrixMatrixMultiply(line[i]:-R,R)
            end do;

            return (R[4,6]+R[4,3]*R[3,6]/(1-R[3,3]))/(1-R[4,4]-R[4,3]*R[3,4]/(1-R[3,3]))
        end if;
    end proc; # etapy

#------------------ Transport functions
    TwissTran:=(line::Element,twiss::Vector)->\
    <line:-R[1,1]^2*twiss[1]-2*line:-R[1,1]*line:-R[1,2]*twiss[2]+line:-R[1,2]^2*twiss[3],\
    -line:-R[2,1]*line:-R[1,1]*twiss[1]+(1+2*line:-R[1,2]*line:-R[2,1])*twiss[2]-line:-R[1,2]*line:-R[2,2]*twiss[3],\
    line:-R[2,1]^2*twiss[1]-2*line:-R[2,2]*line:-R[2,1]*twiss[2]+line:-R[2,2]^2*twiss[3],\
    line:-R[3,3]^2*twiss[4]-2*line:-R[3,3]*line:-R[3,4]*twiss[5]+line:-R[3,4]^2*twiss[6],\
    -line:-R[4,3]*line:-R[3,3]*twiss[4]+(1+2*line:-R[3,4]*line:-R[4,3])*twiss[5]-line:-R[3,4]*line:-R[4,4]*twiss[6],\
    line:-R[4,3]^2*twiss[4]-2*line:-R[4,4]*line:-R[4,3]*twiss[5]+line:-R[4,4]^2*twiss[6]>;

    EtaTran:=(line::Element,disp::Vector)->\
    <line:-R[1,6]+line:-R[1,1]*disp[1]+line:-R[1,2]*disp[2],\
    line:-R[2,6]+line:-R[2,1]*disp[1]+line:-R[2,2]*disp[2],\
    line:-R[3,6]+line:-R[3,3]*disp[3]+line:-R[3,4]*disp[4],\
    line:-R[4,6]+line:-R[4,3]*disp[3]+line:-R[4,4]*disp[4]>;

    HxTran:=(line::Element,twiss::Vector,disp::Vector) ->\
    TwissTran(line,twiss)[1]*EtaTran(line,disp)[2]^2+2*TwissTran(line,twiss)[2]*EtaTran(line,disp)[1]*EtaTran(line,disp)[2] +\
    TwissTran(line,twiss)[3]*EtaTran(line,disp)[1]^2;

#------------------ Tracking of a Beam ----------------------------
    Track:=proc(line::anything,beam::Beam,$) uses LinearAlgebra;
    description "Track a beam ellipse and particles through a BeamLine or ExpandedLine. Returns a Vector of Beams. 1st element is the incoming beam";
    local beamVec:=`if`(type(line,'ExpandedLine'),Vector[row](numelems(line)+1),Vector[row](2)); # the Beam Vector (one loc per line element)
    local Npart:=`if`(member(Nparticles,beam),beam:-Nparticles,0);
    local particleVector:=`if`(Npart>0,table(),0); # the Coordinates Vector (one loc per particle)
    local Sigm:=Matrix(6,6);
    local location,particle,n,loc1,dum;

        BeamParticles:=Npart; # if Elements (like Foil) need this number
        if (Npart > 0) then # track with particles

            if (type(line,'ExpandedLine')) then # track through the pieces of an Expanded line
                beamVec[1]:=Record[beam]();

                for location from 1 to numelems(line) do # loop over elements in line
                    loc1:=location+1;
                    beamVec[loc1]:=Record[beamVec[location]]();
                    dum:=[trackElement(line[location],beamVec[location])];
                    particleVector:=dum[1];Sigm:=dum[2];
                    beamVec[loc1]:-Coordinates:=copy(particleVector);
                    beamVec[loc1]:-Sigma:=Sigm;

                end do; # loop over elements

            elif (type(line,'BeamLine')) then
                beamVec[1]:=Record[beam]();
                beamVec[2]:=Record[beam]();
                dum:=[trackElement(line,beamVec[1])];
                particleVector:=dum[1];Sigm:=dum[2];
                beamVec[2]:-Coordinates:=copy(particleVector);
                beamVec[2]:-Sigma:=Sigm;
            else
                error "First argument not a BeamLine";
            end if; # ExpandedLine

        else

            if (type(line,'BeamLine')) then
                beamVec[1]:=Record[beam]();
                beamVec[2]:=Record[beam]();
                Sigm:=evalf~(Multiply(line:-R,Multiply(beam:-Sigma,Transpose(line:-R))));
                beamVec[2]:-Sigma:=Sigm;
            else
                print("Tracking ExpandedLine without particles not yet implemented");
            end if; # BeamLine

        end if; # track with particles

        return beamVec;
    end proc; # Track

    trackElement:=proc(elemt,beam) uses LinearAlgebra;
    local Sigm,particle,particleVector:=table();
    local f;
        f:=unapply(elemt:-TF(<pv[1],pv[2],pv[3],pv[4],pv[5],pv[6]>),pv); # This makes TF explicit for BeamLines.

        for particle in indices(beam:-Coordinates,'nolist') do # loop over particles
            particleVector[particle]:=evalf~(f(beam:-Coordinates[particle])); # TF will assign a number if particle lost
            if (not type(particleVector[particle],'Vector')) then unassign('particleVector[particle]') end if;
        end do; # loop over particles

        Sigm:=evalf~(Multiply(elemt:-R,Multiply(beam:-Sigma,Transpose(elemt:-R))));
        return particleVector,Sigm;
    end proc; # trackElement

# Our types: we have Elements, BeamLines and Machines
# Note: Have to update these if/when we change any element record.
    ModuleLoad:=proc()
    local idx:=SearchText("Maple",interface(version))+5;
    local vers:=parse(substring(interface(version),idx..idx+5));

# Define our types
        TypeTools:-AddType(Element,'record(l,R)');
        TypeTools:-AddType(BeamLine,'record(l,R,BL)');
        TypeTools:-AddType(ExpandedLine,''Vector''('Element'));
        TypeTools:-AddType(Machine,'record(l,R,Title,BL)');
        TypeTools:-AddType(Beam,'record(Particle)');

# Fill the table with particle parameters [mass(MeV),charge(units),gyromagnetic G]:
        ParticleTable[Electron]:=[0.510998910,-1,0.00115965219];
        ParticleTable[Positron]:=[0.510998910,+1,0.00115965219];
        ParticleTable[Proton]:=[938.272013,+1,1.79284739];
        ParticleTable[Antiproton]:=[938.272013,-1,1.79284739];
        ParticleTable[Pbar]:=[938.272013,-1,1.79284739];
        ParticleTable[Mu]:=[105.6583668,-1,11659208.9E-10];

        unprotect(Spin);
        unprotect('betaxT');
        unprotect('alphaxT');
        unprotect('etaxT');
        unprotect('etapxT');
        unprotect('betayT');
        unprotect('alphayT');
        unprotect('etayT');
        unprotect('etapyT');

# change typesetting to standard for Maple 15
        typesettingOld:=interface(typesetting);

        if (vers>=15 and vers < 16 and typesettingOld='extended') then
            interface(typesetting=standard);
            print("Typesetting set to standard to avoid Maple 15 GUI bug");
        end if; # Maple version

        print("Lattice.mw, Version 1.0.2, 13-Feb-2016");
    end proc; # ModuleLoad

# Revert typesetting change on unload if Maple 15.
    ModuleUnLoad:=proc()
    local idx:=SearchText("Maple",interface(version))+5;
    local vers:=parse(substring(interface(version),idx..idx+5));

        if (vers>=15 and vers < 16 and typesettingOld='extended') then
            if (interface(typesetting)='standard') then
                interface(typesetting='extended');
                print("Typesetting reset to extended")
            end if;

        end if; # Maple version

    end proc; # ModuleUnLoad

#################################################################
#---------------- Synchrotron Integrals
#################################################################

    (* These work for single Elements and for ExpandedLines.

   Notes: Calculations for ExpandedLine debugged (12-Mar-2015)
          Calculations for (dipole-) Element also debugged.
          I4 disagrees with DIMAD for single dipole; likely a DIMAD
          issue as it prints a negative number (for pos. dispersion).
   Uses aux. functions EtaIntf and HIntf local to module Lattice.

   Parameters:
    elm        the element. Has to be type Element or ExpandedLine
    Disp       the dispersion Vector where needed. Column with 4 elements
    Twis       the Twiss Vector where needed (I5). Column with 6 elements
*)

#-----------------I1-----------------------------------------------
    I1:=proc(elm,Disp::Vector)
    description "Evaluate synchrotron integral I1 for elm";
    local etaAvg,etai,ii;

        if (type(elm,'BeamLine')\
            or not (type(elm,'Element')\
                    or type(elm,'ExpandedLine'))) then
            etaAvg:=0;
            error "Not an Element or ExpandedLine";
        elif (type(elm,'ExpandedLine')) then
            etaAvg:=0;
            etai:=copy(Disp);

            for ii from 1 to numelems(elm) do
                etaAvg:=etaAvg+\
                `if`(elm[ii]:-kind=':-Bend',\
                     dispIntf(etai,elm[ii]:-rho,elm[ii]:-a)/elm[ii]:-rho,0);
                etai:=EtaTran(elm[ii],etai);
            end do; # integrate over elements

        elif (elm:-kind=':-Bend') then # only if Dipole
            etaAvg:=dispIntf(Disp,elm:-rho,elm:-a)/elm:-rho;
        else
            etaAvg:=0;
        end if; # select by type

        return etaAvg;
    end proc; # I1
#
#-----------------I2-----------------------------------------------
    I2:=proc(elm)
    description "Evaluate synchrotron integral I2 for elm for length ltot";
    local I21,ii;

        if (type(elm,'BeamLine')\
            or not (type(elm,'Element')\
                    or type(elm,'ExpandedLine'))) then
            I21:=0;
            error "Not an Element or ExpandedLine";
        elif (type(elm,'ExpandedLine')) then
            I21:=0;

            for ii from 1 to numelems(elm) do
                I21:=I21+\
                `if`(elm[ii]:-kind=':-Bend',\
                     elm[ii]:-l/elm[ii]:-rho^2,0);
            end do; # integrate over elements

        elif (elm:-kind=':-Bend') then # only if Dipole
            I21:=elm:-l/elm:-rho^2;
        else I21:=0;
        end if; # selection by elm type

        return I21;
    end proc; # I2
#
#-----------------I3-----------------------------------------------
    I3:=proc(elm)
    description "Evaluate synchrotron integral I3 for elm for length ltot";
    local I31,ii;

        if (type(elm,'BeamLine')\
            or not (type(elm,'Element')\
                    or type(elm,'ExpandedLine'))) then
            I31:=0;
            error "Not an Element or ExpandedLine";
        elif (type(elm,'ExpandedLine')) then
            I31:=0;

            for ii from 1 to numelems(elm) do
                I31:=I31+\
                `if`(elm[ii]:-kind=':-Bend',\
                     elm[ii]:-l/elm[ii]:-rho^3,0);
            end do; # integrate over elements

        elif (elm:-kind=':-Bend') then # only if Dipole
            I31:=elm:-l/elm:-rho^3;
        else I31:=0;
        end if; # selection by elm type

        return I31;
    end proc; # I3
#
#-----------------I4-----------------------------------------------
    I4x:=proc(elm,Disp::Vector)
    description "Evaluate synchrotron integral I4x for elm for length ltot";
    local I41,ii,etai;
        etai:=copy(Disp);
        I41:=0;

        if (type(elm,'BeamLine')\
            or not (type(elm,'Element')\
                    or type(elm,'ExpandedLine'))) then
            error "Not an Element or ExpandedLine";
        elif (type(elm,'ExpandedLine')) then

            for ii from 1 to numelems(elm) do
                I41:=I41+\
                `if`(elm[ii]:-kind=':-Bend',\
                     dispIntf(etai,elm[ii]:-rho,elm[ii]:-a)/elm[ii]:-rho^3,0);
                etai:=EtaTran(elm[ii],etai); # get dispersion @ next element
            end do; # integrate over elements


        elif (elm:-kind=':-Bend') then # only if Dipole
            I41:=dispIntf(etai,elm:-rho,elm:-a)/elm:-rho^3;
        else I41:=0;
        end if; # selection by elm type

        return I41;
    end proc; # I4

#
#-----------------I5-----------------------------------------------
    I5x:=proc(elm,Twis::Vector,Disp::Vector)
    description "Evaluate synchrotron integral I4x for elm for length ltot";
    local I51,ii,etai,twi;
        twi:=copy(Twis);
        etai:=copy(Disp);
        I51:=0;

        if (type(elm,'BeamLine')\
            or not (type(elm,'Element')\
                    or type(elm,'ExpandedLine'))) then
            error "Not an Element or ExpandedLine";
        elif (type(elm,'ExpandedLine')) then

            for ii from 1 to numelems(elm) do
                I51:=I51+\
                `if`(elm[ii]:-kind=':-Bend',\
                     HIntf(twi,etai,elm[ii]:-rho,elm[ii]:-a)/elm[ii]:-rho^3,0);
                etai:=EtaTran(elm[ii],etai); # get dispersion @ next element
                twi:=TwissTran(elm[ii],twi); # gett Twiss fctns @ next element
            end do; # integrate over elements of ExpandedLine


        elif (elm:-kind=':-Bend') then # only if Dipole
            I51:=HIntf(twi,etai,elm:-rho,elm:-a)/elm:-rho^3;
        else I51:=0;
        end if; # selection by elm type

        return I51;
    end proc; # I5

#
    ModuleLoad();
end module;

